# from socialgene.neo4j.neo4j_element import Node, Relationship
# try to use classes already created that inherit from Nodes() or Relationship()
# to do that create a mixin class in a separate file and then import that
